//
// Sample to image postprocessing - painted effect via GLSL programs
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#include    <glut.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "TypeDefs.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "boxes.h"
#include    "FrameBuffer.h"
#include    "GlslProgram.h"

#define MAX_PARTICLES   20000

Vector3D    eye   ( -0.5, -0.5, 1.5 );  // camera position
unsigned    decalMap;                   // decal (diffuse) texture
unsigned    stoneMap;
unsigned    teapotMap;
                                        // various brushes
unsigned    brushMap;
unsigned    dotsBrushMap;
unsigned    normalBrushMap;
unsigned    diagonalBrushMap;
unsigned    linesBrushMap;

float       brushSize = 8;
float       angle     = 0;
float       rot       = 0;
bool        useFilter = true;

Vector2D    particles [MAX_PARTICLES];

GlslProgram 	program;
FrameBuffer     buffer ( 512, 512, FrameBuffer :: depth32 );

void    renderToBuffer ();

void init ()
{
    glClearColor ( 0.0, 0.0, 0.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
    glEnable     ( GL_TEXTURE_2D );
    glDepthFunc  ( GL_LEQUAL     );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void displayBoxes ()
{
    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glRotatef    ( rot, 0, 0, 1 );

    drawBox  ( Vector3D ( -5, -5, 0 ), Vector3D ( 10, 10, 3 ), stoneMap, false );
    drawBox  ( Vector3D ( 3, 2, 0.5 ), Vector3D ( 1,  2,  2 ), decalMap );

    glBindTexture   ( GL_TEXTURE_2D, teapotMap );
    glTranslatef    ( 0.2, 1, 1.5 );
    glRotatef       ( angle * 45.3, 1, 0, 0 );
    glRotatef       ( angle * 57.2, 0, 1, 0 );
    glutSolidTeapot ( 0.3 );

    glPopMatrix     ();
}

void    startOrtho ()
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, 512, 0, 512, -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix

    glDisable ( GL_DEPTH_TEST );
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix

    glEnable ( GL_DEPTH_TEST );
}

void display ()
{
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    renderToBuffer ();

    startOrtho ();

    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_2D, buffer.getColorBuffer ( 0 ) );
    glEnable           ( GL_TEXTURE_2D );

    glActiveTextureARB ( GL_TEXTURE1_ARB );
    glBindTexture      ( GL_TEXTURE_2D, brushMap );

    glActiveTextureARB ( GL_TEXTURE0_ARB );

    glBegin ( GL_QUADS );

        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( 1,   0 );
        glVertex2f   ( 511, 0 );

        glTexCoord2f ( 1, 1 );
        glVertex2f   ( 511, 511 );

        glTexCoord2f ( 0, 1 );
        glVertex2f   ( 0, 511 );

    glEnd   ();

    if ( useFilter )
    {
        program.bind ();

        glEnable    ( GL_BLEND );
        glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

        float   quadratic [] = { 1.0f, 0.0f, 0.0f };

        glEnable              ( GL_POINT_SPRITE_ARB );
        glPointParameterfvARB ( GL_POINT_DISTANCE_ATTENUATION_ARB, quadratic );
        glPointParameterfARB  ( GL_POINT_FADE_THRESHOLD_SIZE_ARB,  20.0f );

        glTexEnvf   ( GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE );
        glPointSize ( brushSize );

        glBegin( GL_POINTS );

                                            // create a random pattern of sprites
        for ( int i = 0; i < MAX_PARTICLES; i++ )
            glVertex2fv ( particles [i] );

        glEnd ();

        glDisable ( GL_BLEND );

        program.unbind ();
    }

    endOrtho ();

    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,    // eye
                    3, 3, 1,                // center
                    0, 0, 1 );              // up
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );

    if ( key == 'f' || key == 'F' )
        useFilter = !useFilter;
    else
    if ( key == '1' )
        brushMap = dotsBrushMap;
    else
    if ( key == '2' )
        brushMap = normalBrushMap;
    else
    if ( key == '3' )
        brushMap = diagonalBrushMap;
    else
    if ( key == '4' )
        brushMap = linesBrushMap;
    else
    if ( key == '+' )
        brushSize ++;
    else
    if ( key == '-' && brushSize >= 1 )
        brushSize--;
}

void    specialKey ( int key, int x, int y )
{
    if ( key == GLUT_KEY_RIGHT )
        rot += 5;
    else
    if ( key == GLUT_KEY_LEFT )
        rot -= 5;

    glutPostRedisplay ();
}

void    animate ()
{
    angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );

    program.bind ();
    program.setUniformFloat ( "time", angle );
    program.unbind ();

    glutPostRedisplay ();
}

void    renderToBuffer ()
{
	buffer.bind ();
	
    glClearColor ( 0, 0, 1, 1 );
    glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    reshape ( buffer.getWidth (), buffer.getHeight () );

    displayBoxes ();

    buffer.unbind ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    glutInitWindowSize  ( 512, 512 );


                                // create window
    glutCreateWindow ( "OpenGL image postprocessing - painted" );

                                // register handlers
    glutDisplayFunc  ( display    );
    glutReshapeFunc  ( reshape    );
    glutKeyboardFunc ( key        );
    glutSpecialFunc  ( specialKey );
    glutIdleFunc     ( animate    );

    init           ();
    initExtensions ();

    if ( !buffer.isSupported () )
    {
        printf ( "FrameBuffer not supported\n" );

        return 1;
    }

	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported.\n" );

		return 1;
	}

    if ( !program.loadShaders ( "brush.vsh", "brush.fsh" ) )
    {
        printf ( "Error loading shaders:\n%s\n", program.getLog ().c_str () );

        return 3;
    }

	unsigned screenMap = buffer.createColorTexture ( GL_RGBA, GL_RGBA8 );
	
	buffer.create ();
	buffer.bind   ();
	
	if ( !buffer.attachColorTexture ( GL_TEXTURE_2D, screenMap, 0 ) )
		printf ( "buffer error with color attachment\n");

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );
					
	buffer.unbind ();

    decalMap  = createTexture2D ( true, "../../Textures/oak.bmp"      );
    stoneMap  = createTexture2D ( true, "../../Textures/block.bmp"    );
    teapotMap = createTexture2D ( true, "../../Textures/Oxidated.jpg" );

    dotsBrushMap     = createTexture2D ( true, "Brushes/dots.tga"           );
    normalBrushMap   = createTexture2D ( true, "Brushes/brush_normal.tga"   );
    diagonalBrushMap = createTexture2D ( true, "Brushes/diagonal.tga"       );
    linesBrushMap    = createTexture2D ( true, "Brushes/diagonal_lines.tga" );
    brushMap         = diagonalBrushMap;

    unsigned map = createTexture2D ( true, "paper.dds" );

    program.bind ();
    program.setTexture ( "mainTex",  0 );
    program.setTexture ( "brushTex", 1 );
    program.unbind ();

                                            // create a random pattern of sprites
    for ( int i = 0; i < MAX_PARTICLES; i++ )
    {
        particles [i].x = (512.0*rand ()) / (float)RAND_MAX;
        particles [i].y = (512.0*rand ()) / (float)RAND_MAX;
    }

    printf ( "Create \"painted\" effect by applying brush\n" );
    printf ( "Keys:\n" );
    printf ( "\tf     turn effect on/off\n" );
    printf ( "\t+     increment brush size by 1.0\n" );
    printf ( "\t-     decrement brush size by 1.0\n" );
    printf ( "\t1     use dots brush      (Brushes/dots.tga)\n" );
    printf ( "\t2     use normal brush    (Brushes/brush_normal.tga)\n" );
    printf ( "\t3     use diagonal brush  (Brushes/diagonal.tga)\n" );
    printf ( "\t4     use lines brush map (Brushes/diagonal_lines.tga)\n" );

    glutMainLoop ();

    return 0;
}
